import concurrent
import itertools
import operator
import sys
import threading
import typing
from collections import Counter
from concurrent.futures import ThreadPoolExecutor

from applitools.common import TestResults, TestResultsSummary, logger
from applitools.common.utils import counted, datetime_utils, iteritems
from applitools.core import EyesRunner

from .helpers import collect_test_results, wait_till_tests_completed
from .resource_cache import ResourceCache

if typing.TYPE_CHECKING:
    from typing import Dict, List, Optional

    from applitools.selenium.visual_grid import RunningTest, VGTask, VisualGridEyes


class VisualGridRunner(EyesRunner):
    def __init__(self, concurrent_sessions=None):
        # type: (Optional[int]) -> None
        super(VisualGridRunner, self).__init__()
        self._all_test_results = {}  # type: Dict[RunningTest, TestResults]

        kwargs = {}
        if sys.version_info[:2] >= (3, 6):
            kwargs["thread_name_prefix"] = "VGR-Executor"

        self.resource_cache = ResourceCache()  # type:ResourceCache
        self.put_cache = ResourceCache()  # type:ResourceCache
        self.all_eyes = []  # type: List[VisualGridEyes]
        self.still_running = True  # type: bool

        self._executor = ThreadPoolExecutor(max_workers=concurrent_sessions, **kwargs)
        self._future_to_task = ResourceCache()  # type:ResourceCache
        thread = threading.Thread(target=self._run, args=())
        thread.setName(self.__class__.__name__)
        thread.daemon = True
        thread.start()
        self._thread = thread

    def __del__(self):
        self._stop()

    def aggregate_result(self, test, test_result):
        # type: (RunningTest, TestResults) -> None
        logger.debug(
            "aggregate_result({}, {}) called".format(test.test_uuid, test_result)
        )
        self._all_test_results[test] = test_result

    def open(self, eyes):
        # type: (VisualGridEyes) -> None
        self.all_eyes.append(eyes)
        logger.debug("VisualGridRunner.open(%s)" % eyes)

    def _run(self):
        logger.debug("VisualGridRunner.run()")
        while self.still_running:
            try:
                task = self._task_queue.pop()
                logger.debug("VisualGridRunner got task %s" % task)
            except IndexError:
                datetime_utils.sleep(1000, msg="Waiting for task")
                continue
            future = self._executor.submit(task)
            self._future_to_task[future] = task

    def _stop(self):
        # type: () -> None
        logger.debug("VisualGridRunner.stop()")
        while sum(r.score for r in self._get_all_running_tests()) > 0:
            datetime_utils.sleep(500, msg="Waiting for finishing tests in stop")
        self.still_running = False
        for future in concurrent.futures.as_completed(self._future_to_task):
            task = self._future_to_task[future]
            try:
                future.result()
            except Exception as exc:
                logger.exception("%r generated an exception: %s" % (task, exc))
            else:
                logger.debug("%s task ran" % task)

        self.put_cache.executor.shutdown()
        self.resource_cache.executor.shutdown()
        self._executor.shutdown()
        self._thread.join()

    def _get_all_test_results_impl(self, should_raise_exception=True):
        # type: (bool) -> TestResultsSummary
        wait_till_tests_completed(self._get_all_running_tests)

        # finish processing of all tasks and shutdown threads
        self._stop()

        all_results = collect_test_results(
            self._all_test_results, should_raise_exception
        )
        return TestResultsSummary(all_results)

    @counted
    def _get_all_running_tests(self):
        # type: ()-> List[RunningTest]
        tests = list(itertools.chain.from_iterable(e.test_list for e in self.all_eyes))
        if not bool(self._get_all_running_tests.calls % 15):
            # print state every 15 call
            counter = Counter(t.state for t in tests)
            logger.info(
                "Current tests states: \n{}".format(
                    "\n".join(["\t{} - {}".format(t, c) for t, c in iteritems(counter)])
                )
            )
        return tests

    def _get_all_running_tests_by_score(self):
        # type: () -> List[RunningTest]
        return sorted(
            self._get_all_running_tests(),
            key=operator.attrgetter("score"),
            reverse=True,
        )

    @property
    def _task_queue(self):
        # type: () -> List[VGTask]
        tests_to_run = self._get_all_running_tests_by_score()
        if tests_to_run:
            test_to_run = tests_to_run[0]
            queue = test_to_run.queue
        else:
            queue = []
        return queue



from .eyes_connector import EyesConnector
from .render_task import RenderTask
from .resource_cache import ResourceCache
from .running_test import RunningTest
from .vg_task import VGTask
from .visual_grid_eyes import VisualGridEyes
from .visual_grid_runner import VisualGridRunner

__all__ = (
    "VGTask",
    "RenderTask",
    "ResourceCache",
    "VisualGridEyes",
    "EyesConnector",
    "VisualGridRunner",
    "RunningTest",
)



from __future__ import unicode_literals

import json
import time
import typing
from abc import abstractmethod
from enum import Enum

import attr

from applitools.common import logger
from applitools.common.utils import ABC, datetime_utils
from applitools.common.utils.json_utils import JsonInclude, to_json
from applitools.selenium import resource

if typing.TYPE_CHECKING:
    from typing import Any, Callable, Dict, List, Optional, Text, Union

    from applitools.selenium import EyesWebDriver

MAX_CHUNK_BYTES_IOS = 10 * 1024 * 1024
MAX_CHUNK_BYTES_GENERIC = 256 * 1024 * 1024
SCRIPT_POLL_INTERVAL_MS = 1000


class DomSnapshotFailure(Exception):
    pass


class DomSnapshotScriptError(DomSnapshotFailure):
    pass


class DomSnapshotTimeout(DomSnapshotFailure):
    def __init__(self, timeout_ms):
        # type: (int) -> None
        super(DomSnapshotTimeout, self).__init__(
            "Dom-Snapshot polling took more than {} ms".format(timeout_ms)
        )


def create_dom_snapshot(
    driver, dont_fetch_resources, skip_resources, timeout_ms=5 * 60 * 1000
):
    # type: (EyesWebDriver, bool, List[Text], int) -> Dict
    is_ie = driver.user_agent.is_internet_explorer
    script_type = DomSnapshotScriptForIE if is_ie else DomSnapshotScriptGeneric
    script = script_type(driver)
    is_ios = "ios" in driver.desired_capabilities.get("platformName", "").lower()
    chunk_byte_length = MAX_CHUNK_BYTES_IOS if is_ios else MAX_CHUNK_BYTES_GENERIC
    return create_dom_snapshot_loop(
        script,
        timeout_ms,
        SCRIPT_POLL_INTERVAL_MS,
        chunk_byte_length,
        dont_fetch_resources=dont_fetch_resources,
        skip_resources=skip_resources,
        serialize_resources=True,
    )


@attr.s
class ProcessPageArgs(object):
    show_logs = attr.ib(
        type=bool, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[bool]
    use_session_cache = attr.ib(
        type=bool, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[bool]
    dont_fetch_resources = attr.ib(
        type=bool, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[bool]
    fetch_timeout = attr.ib(
        type=int, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[int]
    skip_resources = attr.ib(
        type=list, metadata={JsonInclude.NON_NONE: True}
    )  # type: List[Text]
    compress_resources = attr.ib(
        type=bool, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[bool]
    serialize_resources = attr.ib(
        type=bool, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[bool]


@attr.s
class PollResultArgs(object):
    chunk_byte_length = attr.ib(
        type=int, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[int]


class ProcessPageStatus(Enum):
    SUCCESS = "SUCCESS"
    WIP = "WIP"
    ERROR = "ERROR"
    SUCCESS_CHUNKED = "SUCCESS_CHUNKED"


@attr.s
class ProcessPageResult(object):
    status = attr.ib(type=ProcessPageStatus)  # type: ProcessPageStatus
    done = attr.ib(type=bool, default=None)  # type: Optional[bool]
    error = attr.ib(type=str, default=None)  # type: Optional[Text]
    value = attr.ib(default=None)  # type: Optional[Union[Dict,Text]]

    @classmethod
    def from_json(cls, status_json):
        # attr_from_json doesn't support untyped dict fields for now
        status_json = json.loads(status_json)
        return cls(
            status=ProcessPageStatus(status_json["status"]),
            done=status_json.get("done"),
            error=status_json.get("error"),
            value=status_json.get("value"),
        )


class DomSnapshotScript(ABC):
    """Base class for different flavors of dom-snapshot script"""

    @abstractmethod
    def process_page_script_code(self, args):
        # type: (Text) -> Text
        pass

    @abstractmethod
    def poll_result_script_code(self, args):
        # type: (Text) -> Text
        pass

    def __init__(self, driver):
        # type: (EyesWebDriver) -> None
        self._driver = driver

    def run(
        self,
        show_logs=None,  # type: Optional[bool]
        use_session_cache=None,  # type: Optional[bool]
        dont_fetch_resources=None,  # type: Optional[bool]
        fetch_timeout=None,  # type: Optional[int]
        skip_resources=None,  # type: Optional[List[Text]]
        compress_resources=None,  # type: Optional[bool]
        serialize_resources=None,  # type: Optional[bool]
    ):
        # type: (...) -> ProcessPageResult
        return self._run_script(
            locals(), ProcessPageArgs, self.process_page_script_code
        )

    def poll_result(self, chunk_byte_length=None):
        # type: (Optional[bool]) -> ProcessPageResult
        return self._run_script(locals(), PollResultArgs, self.poll_result_script_code)

    def _run_script(self, args, args_type, code_gen_func):
        # type: (Dict[Text, Any], type, Callable[[Text], Text]) -> ProcessPageResult
        args = {k: v for k, v in args.items() if k != "self"}
        code = code_gen_func(to_json(args_type(**args)))
        result_json = self._driver.execute_script(code)
        return ProcessPageResult.from_json(result_json)


class DomSnapshotScriptGeneric(DomSnapshotScript):
    _script_code = resource.get_resource("processPagePoll.js")
    _poll_result_code = resource.get_resource("pollResult.js")

    def process_page_script_code(self, args):
        # type: (Text) -> Text
        return "{} return __processPagePoll({});".format(self._script_code, args)

    def poll_result_script_code(self, args):
        # type: (Text) -> Text
        return "{} return __pollResult({});".format(self._poll_result_code, args)


class DomSnapshotScriptForIE(DomSnapshotScript):
    _script_code = resource.get_resource("processPagePollForIE.js")
    _poll_result_code = resource.get_resource("pollResultForIE.js")

    def process_page_script_code(self, args):
        # type: (Text) -> Text
        return "{} return __processPagePollForIE({});".format(self._script_code, args)

    def poll_result_script_code(self, args):
        # type: (Text) -> Text
        return "{} return __pollResultForIE({});".format(self._poll_result_code, args)


def create_dom_snapshot_loop(
    script, timeout_ms, poll_interval_ms, chunk_byte_length, **script_args
):
    # type: (DomSnapshotScript, int, int, int, **Any) -> Dict
    chunks = []
    deadline = time.time() + datetime_utils.to_sec(timeout_ms)
    result = script.run(**script_args)
    while result.status is ProcessPageStatus.WIP or (
        result.status is ProcessPageStatus.SUCCESS_CHUNKED and not result.done
    ):
        if time.time() > deadline:
            raise DomSnapshotTimeout(timeout_ms)
        result = script.poll_result(chunk_byte_length)
        if result.status is ProcessPageStatus.SUCCESS_CHUNKED:
            logger.info("Snapshot chunk {}, {}B".format(len(chunks), len(result.value)))
            chunks.append(result.value)
        datetime_utils.sleep(poll_interval_ms, "Waiting for the end of DOM extraction")
    if result.status is ProcessPageStatus.SUCCESS:
        return result.value
    elif result.status.SUCCESS_CHUNKED and result.done:
        return json.loads("".join(chunks))
    elif result.status is ProcessPageStatus.ERROR:
        raise DomSnapshotScriptError(result.error)
    else:
        raise DomSnapshotFailure("Unexpected script result", result)



from __future__ import absolute_import, unicode_literals

import json
import re
import threading
import typing as tp
from collections import OrderedDict
from concurrent.futures.thread import ThreadPoolExecutor
from itertools import chain
from typing import Dict, Generator, List, Optional, Text, Union

import attr
import requests
import tinycss2

from applitools.common import EyesError, logger
from applitools.common.utils import (
    datetime_utils,
    is_absolute_url,
    is_url_with_scheme,
    json_utils,
    urljoin,
)
from applitools.common.utils.efficient_string_replace import (
    clean_for_json,
    efficient_string_replace,
)
from applitools.common.utils.json_utils import JsonInclude
from applitools.selenium import eyes_selenium_utils
from applitools.selenium.positioning import ScrollPositionProvider
from applitools.selenium.resource import get_resource

if tp.TYPE_CHECKING:
    from applitools.common.utils.custom_types import AnyWebDriver
    from applitools.selenium.webdriver import EyesWebDriver

__all__ = ("get_full_window_dom",)
_CAPTURE_FRAME_SCRIPT = (
    get_resource("captureDomAndPoll.js") + "return __captureDomAndPoll();"
)
_CAPTURE_FRAME_SCRIPT_FOR_IE = (
    get_resource("captureDomAndPollForIE.js") + "return __captureDomAndPollForIE();"
)
DOM_EXTRACTION_TIMEOUT = 5 * 60 * 1000
CSS_DOWNLOAD_TIMEOUT = 30  # Secs


@attr.s
class Separators(object):
    separator = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    css_start_token = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    css_end_token = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    iframe_start_token = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    iframe_end_token = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text


class CssDownloader(object):
    def __init__(self):
        self.css_start_token = None
        self.css_end_token = None
        self._executor = ThreadPoolExecutor(4)
        self._results = []

    def fetch_css_files(self, base_url, css_start_token, css_end_token, urls):
        #  type: (Text, Text, Text, List[Text]) -> None
        if not is_absolute_url(base_url):
            logger.info("Base URL is not an absolute URL!")
        assert self.css_start_token is None or self.css_start_token == css_start_token
        assert self.css_end_token is None or self.css_end_token == css_end_token
        self.css_start_token = css_start_token
        self.css_end_token = css_end_token
        nodes = [CssNode.create(base_url, url) for url in urls]
        futures = self._executor.map(
            _download_jsonify_node, urls, nodes, timeout=CSS_DOWNLOAD_TIMEOUT
        )
        self._results.append(futures)

    def results(self):
        # type: () -> Dict[Text, Text]
        return {url: data for url, data in chain(*self._results)}

    def __enter__(self):
        self._executor.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._executor.__exit__(exc_type, exc_val, exc_tb)


def _parse_script_result(script_result):
    # type: (Text) -> (Separators, List, List, List)
    missing_css, missing_frames, data = [], [], []
    lines = re.split(r"\r?\n", script_result)
    separators = json_utils.attr_from_json(lines[0], Separators)
    blocks = [missing_css, missing_frames, data]
    block_index = 0
    for line in lines[1:]:
        if separators.separator == line:
            block_index += 1
        else:
            blocks[block_index].append(line)
    logger.info("missing css count: {}".format(len(missing_css)))
    logger.info("missing frames count: {}".format(len(missing_frames)))
    return separators, missing_css, missing_frames, data


def recurse_frames(driver, missing_frames_list, css_downoader):
    # type: (EyesWebDriver, List[Text], CssDownloader) -> Dict[Text, Text]
    frame_data = {}
    switch_to = driver.switch_to
    fc = driver.frame_chain.clone()
    for missing_frame_line in missing_frames_list:
        logger.info("Switching to frame line: {}".format(missing_frame_line))
        original_location = _frame_url(driver)
        try:
            for missing_frame_xpath in missing_frame_line.split(","):
                logger.info("Switching to specific frame: " + missing_frame_xpath)
                frame = driver.find_element_by_xpath(missing_frame_xpath)
                logger.info(
                    "Switched to frame({}) with src ({})".format(
                        missing_frame_xpath, frame.get_attribute("src")
                    )
                )
                switch_to.frame(frame)
            location_after_switch = _frame_url(driver)
            if location_after_switch == original_location:
                logger.info("Switching to frame failed")
                frame_data[missing_frame_line] = ""
                continue
            frame_data[missing_frame_line] = get_frame_dom(driver, css_downoader)
        except Exception:
            logger.exception("Failed to get frame dom")
            frame_data[missing_frame_line] = ""
        switch_to.frames(fc)
    return frame_data


def get_frame_dom(driver, css_downoader):
    # type: (EyesWebDriver, CssDownloader) -> Text
    is_ie = driver.user_agent.is_internet_explorer
    script = _CAPTURE_FRAME_SCRIPT_FOR_IE if is_ie else _CAPTURE_FRAME_SCRIPT
    script_result = get_dom_script_result(
        driver, DOM_EXTRACTION_TIMEOUT, "DomCapture_StopWatch", script
    )
    separators, missing_css, missing_frames, data = _parse_script_result(script_result)

    css_downoader.fetch_css_files(
        _frame_url(driver),
        separators.css_start_token,
        separators.css_end_token,
        missing_css,
    )
    frame_data = recurse_frames(driver, missing_frames, css_downoader)
    return efficient_string_replace(
        separators.iframe_start_token, separators.iframe_end_token, data[0], frame_data
    )


def get_dom(driver):
    # type: (EyesWebDriver) -> Text
    with CssDownloader() as css_downloader, driver.saved_frame_chain():
        dom = get_frame_dom(driver, css_downloader)
        return efficient_string_replace(
            css_downloader.css_start_token,
            css_downloader.css_end_token,
            dom,
            css_downloader.results(),
        )


@datetime_utils.timeit
def get_full_window_dom(driver, return_as_dict=False):
    # type: (EyesWebDriver, bool) -> Union[str, dict]
    current_root_element = eyes_selenium_utils.curr_frame_scroll_root_element(driver)

    with eyes_selenium_utils.get_and_restore_state(
        ScrollPositionProvider(driver, current_root_element)
    ):
        logger.debug("Traverse DOM Tree")
        script_result = get_dom(driver)

    return json.loads(script_result) if return_as_dict else script_result


def get_dom_script_result(driver, dom_extraction_timeout, timer_name, script_for_run):
    # type: (AnyWebDriver, int, Text, Text) -> Dict
    is_check_timer_timeout = []
    script_response = {}
    status = None

    def start_timer():
        def set_timer():
            is_check_timer_timeout.append(True)

        timer = threading.Timer(
            datetime_utils.to_sec(dom_extraction_timeout), set_timer
        )
        timer.daemon = True
        timer.setName(timer_name)
        timer.start()
        return timer

    timer = start_timer()
    while True:
        if status == "SUCCESS" or is_check_timer_timeout:
            del is_check_timer_timeout[:]
            break
        script_result_string = driver.execute_script(script_for_run)
        try:
            script_response = json.loads(
                script_result_string, object_pairs_hook=OrderedDict
            )
            status = script_response.get("status")
        except Exception as e:
            logger.exception(e)
        datetime_utils.sleep(1000, "Waiting for the end of DOM extraction")
    timer.cancel()
    script_result = script_response.get("value")
    if script_result is None or status == "ERROR":
        raise EyesError("Failed to capture script_result")
    return script_result


def _download_jsonify_node(url, node):
    # type: (Text, CssNode) -> (Text, Text)
    return url, clean_for_json(_process_raw_css_node(node, minimize_css=False))


def _process_raw_css_node(node, minimize_css=True):
    # type: (CssNode, bool) -> Text

    @datetime_utils.retry()
    def get_css(url):
        if url.startswith("blob:") or url.startswith("data:"):
            logger.warning("Passing blob URL: {}".format(url))
            return ""
        return requests.get(url, timeout=CSS_DOWNLOAD_TIMEOUT).text.strip()

    def iterate_css_sub_nodes(node, text=None):
        if text is None:
            text = node.text
            if node.text is None:
                text = get_css(node.url)

        for sub_node in _parse_and_serialize_css(node, text, minimize_css):
            if sub_node.url:
                text = get_css(sub_node.url)
                # yield from
                for res in iterate_css_sub_nodes(sub_node, text):
                    yield res
                continue
            yield sub_node.text

    return "".join(iterate_css_sub_nodes(node))


def _parse_and_serialize_css(node, text, minimize=False):
    # type: (CssNode, Text, bool) -> Generator
    def is_import_node(n):
        return n.type == "at-rule" and n.lower_at_keyword == "import"

    stylesheet = tinycss2.parse_stylesheet(
        text, skip_comments=True, skip_whitespace=True
    )
    for style_node in stylesheet:
        if is_import_node(style_node):
            for tag in style_node.prelude:
                if tag.type in ("url", "string"):
                    logger.debug("The node has import")
                    yield CssNode.create_sub_node(parent_node=node, href=tag.value)
            continue

        try:
            if minimize and style_node.content:
                try:
                    # remove whitespaces inside blocks
                    style_node.content = [
                        tok for tok in style_node.content if tok.type != "whitespace"
                    ]
                except AttributeError as e:
                    logger.warning(
                        "Cannot serialize item: {}, cause error: {}".format(
                            style_node, str(e)
                        )
                    )
            serialized = style_node.serialize()
            if minimize:
                serialized = (
                    serialized.replace("\n", "").replace("/**/", " ").replace(" {", "{")
                )

        except TypeError as e:
            logger.warning(str(e))
            continue
        yield CssNode.create_serialized_node(text=serialized)


def _make_url(base_url, value):
    # type: (Text, Text) -> Text
    if is_absolute_url(value) and not is_url_with_scheme(value):  # noqa
        url = urljoin("http://", value)
    else:
        url = urljoin(base_url, value)
    return url


def _frame_url(driver):
    # type: (EyesWebDriver) -> Text
    return driver.execute_script("return document.location.href")


class CssNode(object):
    __slots__ = ("base_url", "url", "text")

    def __init__(self, base_url, url, text):
        # type: (Optional[Text], Optional[Text], Optional[Text]) -> None
        self.base_url = base_url
        self.url = url
        self.text = text

    @classmethod
    def create(cls, base_url, href=None, text=None):
        # type: (Text, Optional[Text], Optional[Text]) -> 'CssNode'
        url = _make_url(base_url, href) if href else None
        return cls(base_url, url, text)

    @classmethod
    def create_sub_node(cls, parent_node, href, text=None):
        # type: ('CssNode', Text, Optional[Text]) -> 'CssNode'
        url = _make_url(parent_node.url, href)
        return cls(parent_node.base_url, url, text)

    @classmethod
    def create_serialized_node(cls, text):
        # type: (Text) -> 'CssNode'
        return cls(base_url=None, url=None, text=text)



import copy
import typing as tp

import attr

from applitools.common import EyesError, Point, RectangleSize, logger

from . import eyes_selenium_utils

if tp.TYPE_CHECKING:
    from typing import Optional

    from applitools.common.utils.custom_types import AnyWebDriver

    from .webelement import EyesWebElement

__all__ = ("Frame", "FrameChain")


@attr.s(slots=True)
class Frame(object):
    """
    Encapsulates a frame/iframe. This is a generic type class,
    and it's actual type is determined by the reference used by the user in
    order to switch into the frame.
    :param reference: The web element for the frame, used as a reference to
                      switch  into the frame.
    :param location: The location of the frame within the current frame.
    :param outer_size: The frame element outerSize (i.e., the outer_size of the
                       frame on the screen, not the internal document outer_size).
    :param inner_size: The frame element inner outerSize (i.e., the outer_size
                       of the frame actual outer_size, without borders).
    :param parent_scroll_position: The scroll location of the frame.
    :param scroll_root_element: The element used for content scrolling within frame.
    """

    reference = attr.ib()  # type: EyesWebElement
    location = attr.ib()  # type: Point
    outer_size = attr.ib()  # type: RectangleSize
    inner_size = attr.ib()  # type: RectangleSize
    parent_scroll_position = attr.ib()
    scroll_root_element = attr.ib()  # type: EyesWebElement
    original_overflow = attr.ib(default=None)

    def return_to_original_overflow(self, driver):
        # type: (AnyWebDriver) -> None
        root_element = eyes_selenium_utils.get_underlying_webelement(
            self._scroll_root_element(driver)
        )
        logger.debug(
            "returning overflow of element to its original value: {}".format(
                root_element
            )
        )
        driver.execute_script(
            "arguments[0].style.overflow='%s'" % self.original_overflow, root_element
        )

    def _scroll_root_element(self, driver):
        scroll_root = self.scroll_root_element
        if scroll_root is None or not scroll_root.is_attached_to_page:
            logger.debug("no scroll root element. selecting default.")
            scroll_root = driver.find_element_by_tag_name("html")
        return scroll_root

    def hide_scrollbars(self, driver):
        scroll_root_element = eyes_selenium_utils.get_underlying_webelement(
            self._scroll_root_element(driver)
        )
        self.original_overflow = driver.execute_script(
            "var origOF = arguments[0].style.overflow; arguments["
            "0].style.overflow='hidden'; return origOF;",
            scroll_root_element,
        )
        return self.original_overflow


@attr.s(slots=True, init=False, eq=False)
class FrameChain(tp.Sequence[Frame]):
    _frames = attr.ib()

    def __init__(self, frame_chain=None):
        # type: (Optional[FrameChain]) -> None
        self._frames = []
        if frame_chain is not None:
            assert isinstance(frame_chain, FrameChain), "Must be a FrameChain"
            self._frames = copy.copy(frame_chain._frames)

    def __iter__(self):
        return iter(self._frames)

    def __getitem__(self, item):
        return self._frames[item]

    def __delitem__(self, item):
        del self._frames[item]

    def __len__(self):
        # type: () -> int
        return len(self._frames)

    def __str__(self):
        return "FrameChain with {} frames".format(len(self))

    def __eq__(self, other):
        cl1, cl2 = len(self._frames), len(other)
        if cl1 != cl2:
            return False
        return all(self._frames[i].id_ == other[i].id_ for i in range(cl1))

    def clear(self):
        # type: () -> None
        del self._frames[:]

    @property
    def size(self):
        # type: () -> int
        return len(self)

    @property
    def peek(self):
        # type: () -> Optional[Frame]
        if self._frames:
            return self[-1]
        return None

    def push(self, frame):
        # type: (Frame) -> None
        assert isinstance(frame, Frame), "frame must be instance of Frame!"
        self._frames.append(frame)

    def pop(self):
        # type: () -> Frame
        return self._frames.pop()

    @property
    def current_frame_offset(self):
        # type: () -> Point
        location = Point.ZERO()
        for frame in self:
            location = location.offset(frame.location)
        return location

    @property
    def current_frame_size(self):
        # type: () -> RectangleSize
        return self.peek.inner_size

    @property
    def current_frame_inner_size(self):
        # type: () -> RectangleSize
        return self.peek.outer_size

    @property
    def default_content_scroll_position(self):
        # type: () -> tp.Union[Point, EyesError]
        if len(self) == 0:
            raise EyesError("No frames in frame chain")
        result = self[0].parent_scroll_position
        return Point(result.x, result.y)

    def clone(self):
        # type: () -> FrameChain
        return FrameChain(self)
